/******************************************************************************
**
** FILE NAME    : ifx_ppa_directpath.c
** PROJECT      : 
** MODULES      : PPA Directpath
**
** DATE         : 11 FEB 2008
** AUTHOR       : Santosh Gowda
** DESCRIPTION  : PPA Direct path implementation
** COPYRIGHT    :   Copyright (c) 2006
**          Infineon Technologies AG
**          Am Campeon 1-12, 85579 Neubiberg, Germany
**
**   Any use of this software is subject to the conclusion of a respective
**   License agreement. Without such a License agreement no rights to the
**   software are granted
**
** HISTORY
** $Date        $Author         $Comment
** 11 FEB 2008  Santosh Gowda   Initiate Version
*******************************************************************************/



/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>
#include <linux/in.h>
#include <net/sock.h>
#include <net/ip_vs.h>
#include <asm/time.h>

#include <linux/trace.h>

/*
 *  Chip Specific Head File
 */
#include <asm/amazon_se/ifx_ppa_directpath.h>
#include <asm/amazon_se/ifx_ppa_hook.h>

#define ENABLE_DEBUG_PRINT                    0

/*
 * ####################################
 *           Extern Variable
 * ####################################
 */


/* Packet counters */
 int g_directpath_dbg = 0;
 int g_directpath_tx_pkts = 0;
 int g_directpath_rx_pkts = 0;
 int g_directpath_tx_nonaccel = 0;
 int g_directpath_rx_nonaccel = 0;

#if defined(ENABLE_DEBUG_PRINT)
  #define dbg(format, arg...)    do { if ( g_directpath_dbg ) printk(KERN_WARNING __FILE__ ":%d:%s: " format "\n", __LINE__, __FUNCTION__, ##arg); } while ( 0 )
#elif !defined(dbg)
  #define dbg(format, arg...)
#endif

/* Dump skb */
// #define DUMP_SKB
#undef DUMP_SKB

#ifdef DUMP_SKB
/*
 * dump skb data
 */
static inline void
dump_my_skb (u32 len, char *data)
{
	int i;
	printk("Passed Len [%d]\n", len);
	for (i = 0; i < len; i++) {
		printk ("%2.2x ", (u8) (data[i]));
		if (i % 16 == 15)
			printk ("\n");
	}
	printk ("\n");
}
#endif


/*
 * ####################################
 *           Global Variable
 * ####################################
 */

static PPA_DIRECTPATH_INFO 	gPpaDpathInfo[PPA_MAX_DEV];
static uint32_t 		g_max_ifid;
static uint32_t 		g_min_ifid;
static PPA_LOCK 		g_directpath_info_lock;

/*
 * Table for mapping a Interface bit mask (with one bit per interface level)
 * into the number of the highest priority set bit.
 */
static uint8_t IfMask2IfId[] = {
	0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, /* 00 - 0f */
	4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, /* 10 - 1f */
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 20 - 2f */
	5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, /* 30 - 3f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 40 - 4f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 50 - 5f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 60 - 6f */
	6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, /* 70 - 7f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 80 - 8f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* 90 - 9f */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* a0 - af */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* b0 - bf */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* c0 - cf */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* d0 - df */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, /* e0 - ef */
	7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7  /* f0 - ff */
};


/*
 * Get free slot in gPpaDpathInfo
 */
int32_t ifx_ppa_get_ifid ()
{
	uint32_t index = 0;

	ppa_lock_get (&g_directpath_info_lock);
	for (index = g_min_ifid; index <= g_max_ifid; index++) {
		if (gPpaDpathInfo[index].occupied) {
			continue;
		}
		ppa_lock_release (&g_directpath_info_lock);
		return index;
	}
	ppa_lock_release (&g_directpath_info_lock);
	return -1;
}

/*
 * Map PPA interface id to Protocol stack interface structure
 */

PPA_NETIF *ifx_ppa_get_netif_for_ifid(uint32_t if_id)
{
	if (if_id <= 2) {
		return NULL;
	}
	return (gPpaDpathInfo[if_id].pDev);
}

/*
 * Map Protocol stack interface structure to PPA interface id
 * Also checks if given device exist in Directpath 
 */

int32_t ifx_ppa_get_ifid_for_netif (PPA_NETIF *dev)
{
	uint32_t index = g_min_ifid;

	for (; index < g_max_ifid; index++) {
		/* true if dev->ifindex is matching */
		if (ppa_is_netif_equal(dev, gPpaDpathInfo[index].pDev)) {
			return index;
		}
	}
	return -1;
}

/*
 * called by PPE to send pkt to a CPU bound Directpath interface.
 */

int32_t ifx_ppa_directpath_rx (PPA_BUF *ppa_buf, int32_t len, uint32_t rx_if_id, uint32_t tx_if_mask, void *ppe_header)
{
  	PPA_NETIF *out_dev = NULL;
  	PPA_NETIF *in_dev = NULL;
	uint32_t  ifmask = 0, if_id = 0;
	int32_t ret = IFX_SUCCESS;

	if (unlikely(tx_if_mask == 0)) { 
		goto drop;
	}
	ifmask = tx_if_mask;
	dbg ("ifmask:[%d]", ifmask);

	while (ifmask) {

		if_id = IfMask2IfId[ifmask];
		dbg ("if_id:[%d]", if_id);
		/* Process the if_id if its directpath registered */
		out_dev = ifx_ppa_get_netif_for_ifid (if_id);
		if ((unlikely(out_dev == NULL))) {
			dbg("out_dev is NULL");
			goto drop;
		}
		ppa_lock_get (&g_directpath_info_lock);
			
		if ((unlikely(gPpaDpathInfo[if_id].pDirectpathCb->rx_fn == NULL))) {
			dbg("no Rx callback");
			ppa_lock_release (&g_directpath_info_lock);
			goto drop;
		}
		/* set outgoing dev */
		ppa_buf->dev = out_dev;

		dbg ("PPA --> CPU");
		/* send pkt to corresponding CPU bound device driver */
		gPpaDpathInfo[if_id].pDirectpathCb->rx_fn (in_dev, out_dev, ppa_buf, len);

		gPpaDpathInfo[if_id].rx_pkts++;	
		g_directpath_rx_pkts++;

		ppa_lock_release (&g_directpath_info_lock);
		/* moving to next set-bit */
		ifmask &=~ (1 << (if_id));
	}
	return ret;	

drop:
	dbg ("drop: dropping pkt if_id:[%d]", if_id);
        ppa_kfree_skb(ppa_buf);
	return IFX_FAILURE;
}

/*
 * called by any CPU bound Directpath interface to send pkt to PPE.
 */

int32_t ifx_ppa_directpath_tx (uint32_t rx_if_id, PPA_BUF *ppa_buf, int32_t len, uint32_t flags)
{
	struct session_list_item *p_item = NULL;
	struct bridging_session_list_item *p_br_item = NULL;
	PPA_SESSION *p_session = NULL;

	uint32_t mc_group = 0;
	uint8_t mac_addr[PPA_ETH_ALEN] = {0};
	uint8_t *p_olddata=NULL;
	uint32_t tx_dma_qid = 0;
	int32_t ret = IFX_SUCCESS;

	if (unlikely(ppa_buf == NULL)) {
		dbg ("ppa_buf is NULL !!");
		return IFX_FAILURE;
	}

	ppa_lock_get (&g_directpath_info_lock);
	/* Check if rxif is of Ethernet type */
	if (gPpaDpathInfo[rx_if_id].flags & PPA_F_DIRECTPATH_ETH_IF) {
		/* reach the IP header */
		p_olddata=ppa_get_ppa_buf_data_ptr(ppa_buf);
		/* Check if IP pkt */
		if (ppa_check_is_ip_pkt(p_olddata - ETH_HLEN)== 0) {
			goto drop;
		}
	}
	ppa_set_nh_in_ip_pkt(ppa_buf, p_olddata);
	/* No accel for non-TCP/UDP, broadcast, fragment pkt types */
        ret = ppa_is_pkt_type_supported_for_accel (ppa_buf);
	if (unlikely(ret < 0)) {
		dbg ("no support for accel");
		goto drop;
	}
	/* Check if rxif is of Ethernet type */
	if (gPpaDpathInfo[rx_if_id].flags & PPA_F_DIRECTPATH_ETH_IF) {
		/* rollback to the ETH header */
		if (p_olddata && ppa_set_ppa_buf_data_ptr(ppa_buf, 
			p_olddata, ETH_HLEN ) < 0) {
			dbg("Not an IP packet!\n");
			goto drop; /* Not a IP Packet */
		}
		len += ETH_HLEN; /* Account for Eth hdr again */
	}
	if (unlikely(p_olddata == NULL)) {
		/* Add dummy Ethernet header. Needs to be removed on Rx
		 * if pkt is looped back by PPE due to no accel entry */
		/* TODO: ... */
		dbg("Can't handle today!\n");
		goto drop;
	}

	/* call PPE driver - CPU to PPE Tx fxn */
	dbg ("CPU --> PPA\n");
	/* tx_dma_qid is reserved for now */
	ret = ppa_send_pkt_to_accel_hw (ppa_buf, rx_if_id, tx_dma_qid, len);
	if (unlikely(ret < 0)) {
		/* TODO: If Tx fxn fails, apply flow control using PPA Tx stop hook fxn */ 
		dbg("Apply flow control!\n");
		goto apply_flow_control;
	}
	else {
		gPpaDpathInfo[rx_if_id].tx_pkts++;	
		g_directpath_tx_pkts++;
	}
	ppa_lock_release (&g_directpath_info_lock);
	return ret;

apply_flow_control: 
	gPpaDpathInfo[rx_if_id].tx_drop++;
	ppa_lock_release (&g_directpath_info_lock);
	return ret;

drop:
	dbg ("taking normal path\n");
	gPpaDpathInfo[rx_if_id].tx_non_accel++;	
	g_directpath_tx_nonaccel++;
	ppa_lock_release (&g_directpath_info_lock);
        /* take normal path to help create session */
	netif_rx (ppa_buf);
	return IFX_SUCCESS;	
	
}

int32_t ifx_ppa_directpath_pause_tx (uint32_t if_id, PPA_FP_STOP_TX_FN stop_tx_fn, uint32_t flags)
{
	int32_t ret = IFX_SUCCESS;

	return ret;
}

int32_t ifx_ppa_directpath_resume_tx (uint32_t if_id, PPA_FP_RESTART_TX_FN start_tx_fn, uint32_t flags)
{
	int32_t ret = IFX_SUCCESS;

	return ret;
}

int32_t ifx_ppa_directpath_pause_rx (uint32_t if_id, uint32_t flags)
{
	int32_t ret = IFX_SUCCESS;
	return ret;

}


int32_t ifx_ppa_directpath_resume_rx (uint32_t if_id, uint32_t flags)
{
	int32_t ret = IFX_SUCCESS;
	return ret;
}


int32_t ifx_ppa_directpath_register_dev(uint32_t *if_id, PPA_NETIF *dev, PPA_DIRECTPATH_CB *pDirectpathCb, uint32_t flags)
{
	uint32_t index = 0;
    	PPA_IFINFO ifinfo;
    	PPA_IFNAME *ifname = NULL;
    	int f_is_lanif = 1;

	if (flags < 0 || dev == NULL) {
		goto fail;
	}

	/* Register */ 
	if (flags == 0 || flags == PPA_F_DIRECTPATH_ETH_IF || flags == PPA_F_DIRECTPATH_CORE1) {
		if (pDirectpathCb == NULL) {
			goto fail;
		}
		index = ifx_ppa_get_ifid ();
		if (index < 0) {
			goto fail;
		}
        	ppa_memset(&ifinfo, 0, sizeof(ifinfo));
		ifinfo.ifname = (PPA_IFINFO *) ppa_malloc (sizeof(PPA_IF_NAME_SIZE));
		if (ifinfo.ifname == NULL) {
			dbg ("malloc failed for ifinfo.ifname");
			goto fail;
		}
            	ifname = ppa_get_netif_name(dev);
            	if ( ifname == NULL) {
			dbg ("ifname not valid !!");
			goto fail;
		}
        	strncpy(ifinfo.ifname, ifname, PPA_IF_NAME_SIZE);
		ifinfo.ifname[PPA_IF_NAME_SIZE-1]='\0';
		dbg ("ifname:[%s]", ifinfo.ifname);
            	ifinfo.if_flags = 0;
            	ifx_ppa_add_itf(&ifinfo, f_is_lanif);
		ppa_set_phys_ifname_for_ifid(index, ifname);

		ppa_lock_get (&g_directpath_info_lock);
		gPpaDpathInfo[index].pDev = dev;
		gPpaDpathInfo[index].occupied = IFX_TRUE;
		gPpaDpathInfo[index].flags = flags;
		gPpaDpathInfo[index].pDirectpathCb = pDirectpathCb;
		ppa_lock_release (&g_directpath_info_lock);
		*if_id = index; // return ifid to caller
		g_max_ifid++;

	}
    	else if ( (flags && PPA_F_DIRECTPATH_DEREGISTER) ) {
		if (*if_id <= 2) {
			dbg ("Invalid IfId!!\n");
			goto fail;
		}
		index = *if_id;
		ppa_lock_get (&g_directpath_info_lock);
		gPpaDpathInfo[index].occupied = IFX_FALSE;
		gPpaDpathInfo[index].pDev = NULL;
		gPpaDpathInfo[index].pDirectpathCb = NULL; 
		ppa_set_phys_ifname_for_ifid(index, NULL); /* reset if_id ifname */
		ppa_lock_release (&g_directpath_info_lock);
		g_max_ifid--;
	}      
	return IFX_SUCCESS;

fail:
	return IFX_FAILURE;
}


int32_t ifx_ppa_directpath_init(uint32_t flags)
{

	uint32_t index = 0;

	
	g_min_ifid = 3;
	g_max_ifid = 3;
    	ppa_lock_init(&g_directpath_info_lock);
	for (index = 0; index < PPA_MAX_DEV; index++) {
		gPpaDpathInfo[index].pDev = (PPA_NETIF *) ppa_malloc (sizeof(PPA_NETIF));
		if (!gPpaDpathInfo[index].pDev) {
			dbg ("ppa_malloc failed\n");
			return IFX_FAILURE;
		}

		gPpaDpathInfo[index].pDirectpathCb = (PPA_DIRECTPATH_CB *) ppa_malloc (sizeof(PPA_DIRECTPATH_CB));
		if (!gPpaDpathInfo[index].pDirectpathCb) {
			dbg ("ppa_malloc failed\n");
			return IFX_FAILURE;
		}
		/* init callback fxns */
		gPpaDpathInfo[index].pDirectpathCb->stop_tx_fn = NULL;
		gPpaDpathInfo[index].pDirectpathCb->start_tx_fn = NULL;
		gPpaDpathInfo[index].pDirectpathCb->rx_fn = NULL;
		/* ifid 0,1,2 are reserved for non-cpu bound interfaces */
		if (index <= 2) {
			gPpaDpathInfo[index].occupied = IFX_TRUE;
			continue;
		}
		gPpaDpathInfo[index].occupied = IFX_FALSE;
	}

   	ppa_hook_directpath_register_dev_fn  = ifx_ppa_directpath_register_dev;
	ppa_hook_directpath_send_fn = ifx_ppa_directpath_tx;
        ppa_hook_rx_pkt_fn = ifx_ppa_directpath_rx; 
        ppa_hook_directpath_rx_restart_fn = ifx_ppa_directpath_resume_rx;
	ppa_hook_directpath_rx_stop_fn = ifx_ppa_directpath_pause_rx;
 
        ppa_hook_get_netif_for_ppa_ifid_fn = ifx_ppa_get_netif_for_ifid; 
	ppa_hook_get_ppa_ifid_for_netif_fn = ifx_ppa_get_ifid_for_netif; 

	//g_directpath_enable = IFX_TRUE;
   	printk ("ppa_directpath_init - succeeded\n");
   	return IFX_SUCCESS;
}

int32_t ifx_ppa_directpath_exit(uint32_t flags)
{
	uint32_t index = 0;

	for (index = 0; index < PPA_MAX_DEV; index++) {
		if (gPpaDpathInfo[index].pDev) {
			ppa_free (gPpaDpathInfo[index].pDev);
		}
		gPpaDpathInfo[index].pDev = NULL;

		if (gPpaDpathInfo[index].pDirectpathCb) {
			ppa_free (gPpaDpathInfo[index].pDirectpathCb);
		}
		gPpaDpathInfo[index].pDirectpathCb = NULL;
		gPpaDpathInfo[index].occupied = IFX_FALSE;

	}

   	ppa_hook_directpath_register_dev_fn = NULL;
	ppa_hook_directpath_send_fn = NULL;
        ppa_hook_rx_pkt_fn = NULL; 
        ppa_hook_directpath_rx_restart_fn = NULL;
	ppa_hook_directpath_rx_stop_fn = NULL;

        ppa_hook_get_netif_for_ppa_ifid_fn = NULL; 
	ppa_hook_get_ppa_ifid_for_netif_fn = NULL; 

    	ppa_lock_destroy(&g_directpath_info_lock);
	//g_directpath_enable = IFX_FALSE;
   	printk ("ppa_directpath_exit - done\n");
   	return IFX_SUCCESS;
}

EXPORT_SYMBOL (ifx_ppa_directpath_init);
EXPORT_SYMBOL (ifx_ppa_directpath_exit);

